/*
Bullet Continuous Collision Detection and Physics Library
Copyright (c) 2015 Google Inc. http://bulletphysics.org

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from the use of this software.
Permission is granted to anyone to use this software for any purpose, 
including commercial applications, and to alter it and redistribute it freely, 
subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
*/


#include "Problem1.h"

#include "btBulletDynamicsCommon.h"
#define ARRAY_SIZE_Y 5
#define ARRAY_SIZE_X 5
#define ARRAY_SIZE_Z 5

#include "LinearMath/btVector3.h"
#include "LinearMath/btAlignedObjectArray.h"

#include "BulletDynamics/ConstraintSolver/btGeneric6DofConstraint.h"

#include "../CommonInterfaces/CommonRigidBodyBase.h"
#include "../RenderingExamples/TimeSeriesCanvas.h"


struct Problem1InternalData{
	btRigidBody *bodyFixed;
	btRigidBody *bodyPlatform;
	btGeneric6DofConstraint *constraint;
	btDefaultMotionState *motionState;
	btJointFeedback *jointFeedback;
	TimeSeriesCanvas *canvasTimeSeries;
	float dt;
	unsigned int frameID;
	
	Problem1InternalData() : dt( 1.0 / 60.0 ), frameID( 0 ){ }
};

struct Problem1 : public CommonRigidBodyBase{
	struct Problem1InternalData* data;
	
	Problem1( struct GUIHelperInterface *helper );
	virtual ~Problem1();
	virtual void initPhysics();
	virtual void renderScene();
	virtual void stepSimulation( float deltaTime );
	void resetCamera();
};

Problem1::Problem1( struct GUIHelperInterface *helper ) : CommonRigidBodyBase( helper ){
	data = new Problem1InternalData;
	data->canvasTimeSeries = new TimeSeriesCanvas(helper->get2dCanvasInterface(), 256, 256, "Constraint Status");
	data->canvasTimeSeries ->setupTimeSeries( 20, 100, 0 );
	data->canvasTimeSeries->addDataSource( "Joint Force [N]", 255, 0, 0 );
	data->canvasTimeSeries->addDataSource( "Velocity [m/s]", 0, 0, 255 );
	data->canvasTimeSeries->addDataSource( "Position [m]", 0, 255, 0 );
}

Problem1::~Problem1(){
	exitPhysics();
	
	delete data->jointFeedback;
	data->jointFeedback = 0;
	delete data->canvasTimeSeries;
	data->canvasTimeSeries = 0;
	
	delete data;
}

void Problem1::initPhysics(){
	m_guiHelper->setUpAxis(1);
	
	m_collisionConfiguration = new btDefaultCollisionConfiguration();
	m_dispatcher = new btCollisionDispatcher( m_collisionConfiguration );
	btVector3 worldAabbMin( -10000.0, -10000.0, -10000.0 );
	btVector3 worldAabbMax( 10000.0, 10000.0, 10000.0 );
	m_broadphase = new btAxisSweep3( worldAabbMin, worldAabbMax );
	
	m_solver = new btSequentialImpulseConstraintSolver;
	//m_solver = new btNNCGConstraintSolver;
	
	m_dynamicsWorld = new btDiscreteDynamicsWorld( m_dispatcher, m_broadphase, m_solver, m_collisionConfiguration );
	m_dynamicsWorld->getDispatchInfo().m_useContinuous = true;
	m_guiHelper->createPhysicsDebugDrawer( m_dynamicsWorld );
	
	if( m_dynamicsWorld->getDebugDrawer() ){
		m_dynamicsWorld->getDebugDrawer()->setDebugMode( btIDebugDraw::DBG_DrawWireframe + btIDebugDraw::DBG_DrawContactPoints );
	}
	
	m_dynamicsWorld->getSolverInfo().m_numIterations = 20; //100;
	
	// ground shape
	{
		btCollisionShape * const shape = new btBoxShape( btVector3( 200.0, 5.0, 200.0 ) );
		btTransform transform;
		transform.setIdentity();
		transform.setOrigin( btVector3( 0.0, -5.0, 0.0 ) );
		btCollisionObject * const body = new btCollisionObject();
		body->setCollisionShape( shape );
		body->setWorldTransform( transform );
		m_dynamicsWorld->addCollisionObject( body );
	}
	
	// static rigid body to attach platform to
	{
		btCollisionShape * const shape = new btBoxShape( btVector3( 0.1, 0.1, 0.1 ) );
		btTransform transform;
		transform.setIdentity();
		data->bodyFixed = new btRigidBody( 0.0, NULL, shape, btVector3() );
		data->bodyFixed->setCollisionShape( shape );
		data->bodyFixed->setWorldTransform( transform );
		m_dynamicsWorld->addCollisionObject( data->bodyFixed );
	}
	
	{
		// create platform
		btCollisionShape * const shape = new btBoxShape( btVector3( 0.85, 0.04998, 0.85 ) );
		btTransform platformTransform;
		platformTransform.setIdentity();
		platformTransform.setOrigin( btVector3( 0.0, 2.0, 0.0 ) );
		
		btScalar mass = 1.0;
		btVector3 localInertia( 0.0, 0.0, 0.0 );
		shape->calculateLocalInertia( mass, localInertia );
		
		data->motionState = new btDefaultMotionState( platformTransform );
		data->bodyPlatform = new btRigidBody( mass, data->motionState, shape, localInertia );
		data->bodyPlatform->setActivationState( DISABLE_DEACTIVATION );
		m_dynamicsWorld->addRigidBody( data->bodyPlatform );
		
		// platform constraint. allows moving up and down between two limits.
		// for this problem show-case the limits are identical. in this configuration
		// the constraint equals a fixed constraint. the problem is visible in the
		// joint feedback data and body linear velocity. at the beginning there is
		// a linear velocity for up to 5 frames although the platform should at rest.
		// also there is no correction force for up to 3 frames then suddenly a huge
		// one of 22. Then after frame 5 the constraint hickups seem to go away and
		// it starts fixing the error introducted by the initial hickup and converges
		// to the stable situation. the initial force explosion is though enough to
		// break the test case of "static joint friction" (limits lifted if force
		// exceeds for example 15 [N].
		// 
		// example printf from a run of the demo
		/*
		appliedForce=0 linearVelocity=-0.166667 position=1.99866
		appliedForce=0 linearVelocity=-0.166667 position=1.99813
		appliedForce=0 linearVelocity=-0.166667 position=1.99754
		appliedForce=22 linearVelocity=0.0333338 position=1.99738
		appliedForce=22 linearVelocity=0.0333338 position=1.99751
		appliedForce=9.60003 linearVelocity=0.0266676 position=1.99797
		appliedForce=9.68002 linearVelocity=0.0213346 position=1.99838
		appliedForce=9.74397 linearVelocity=0.0170674 position=1.9987
		appliedForce=9.79521 linearVelocity=0.0136542 position=1.99896
		appliedForce=9.83615 linearVelocity=0.0109234 position=1.99917
		appliedForce=9.86894 linearVelocity=0.00873899 position=1.99934
		appliedForce=9.89511 linearVelocity=0.00699091 position=1.99947
		appliedForce=9.91614 linearVelocity=0.0055933 position=1.99958
		appliedForce=9.93288 linearVelocity=0.00447464 position=1.99966
		appliedForce=9.94627 linearVelocity=0.00357914 position=1.99973
		appliedForce=9.95708 linearVelocity=0.00286388 position=1.99978
		appliedForce=9.96567 linearVelocity=0.00229168 position=1.99983
		appliedForce=9.97253 linearVelocity=0.00183392 position=1.99986
		appliedForce=9.97803 linearVelocity=0.0014677 position=1.99989
		appliedForce=9.9824 linearVelocity=0.00117445 position=1.99991
		appliedForce=9.98592 linearVelocity=0.000939846 position=1.99993
		appliedForce=9.98876 linearVelocity=0.000752449 position=1.99994
		appliedForce=9.99099 linearVelocity=0.000602245 position=1.99995
		appliedForce=9.99279 linearVelocity=0.000482082 position=1.99996
		appliedForce=9.99425 linearVelocity=0.000386238 position=1.99997
		appliedForce=9.99537 linearVelocity=0.00030899 position=1.99998
		appliedForce=9.99631 linearVelocity=0.000247478 position=1.99998
		appliedForce=9.997 linearVelocity=0.000197411 position=1.99999
		appliedForce=9.9976 linearVelocity=0.000157356 position=1.99999
		appliedForce=9.99811 linearVelocity=0.000125885 position=1.99999
		appliedForce=9.99846 linearVelocity=0.000100136 position=1.99999
		appliedForce=9.9988 linearVelocity=8.01086e-05 position=1.99999
		appliedForce=9.99906 linearVelocity=6.4373e-05 position=2
		appliedForce=9.99923 linearVelocity=5.14984e-05 position=2
		appliedForce=9.9994 linearVelocity=4.14848e-05 position=2
		appliedForce=9.99949 linearVelocity=3.29018e-05 position=2
		appliedForce=9.99957 linearVelocity=2.57492e-05 position=2
		appliedForce=9.99966 linearVelocity=2.00272e-05 position=2
		appliedForce=9.99974 linearVelocity=1.57356e-05 position=2
		appliedForce=9.99983 linearVelocity=1.28746e-05 position=2
		appliedForce=9.99983 linearVelocity=1.00136e-05 position=2
		appliedForce=9.99991 linearVelocity=8.58307e-06 position=2
		appliedForce=9.99991 linearVelocity=7.15256e-06 position=2
		appliedForce=9.99991 linearVelocity=5.72205e-06 position=2
		appliedForce=9.99991 linearVelocity=4.29153e-06 position=2
		appliedForce=9.99991 linearVelocity=2.86102e-06 position=2
		appliedForce=10 linearVelocity=2.86102e-06 position=2
		appliedForce=10 linearVelocity=2.86102e-06 position=2
		appliedForce=10 linearVelocity=2.86102e-06 position=2
		*/
		
		btTransform localA;
		localA.setIdentity();
		localA.getOrigin() = btVector3( 0.0, 0.0, 0.0 );
		
		btTransform localB;
		localB.setIdentity();
		localB.getOrigin() = btVector3( 0.0, 2.0, 0.0 );
		
		btGeneric6DofConstraint * const constraint =  new btGeneric6DofConstraint(
			*data->bodyPlatform, *data->bodyFixed, localA, localB, true );
		int i;
		for( i=0; i<6; i++ ){
			constraint->setLimit( i, 0.0, 0.0 );
			constraint->setParam( BT_CONSTRAINT_CFM, 0.0, i ); // default 0.0f
			constraint->setParam( BT_CONSTRAINT_STOP_CFM, 0.0, i ); // default 0.0f
			constraint->setParam( BT_CONSTRAINT_STOP_ERP, 0.2, i ); // default 0.2f
		}
		constraint->getTranslationalLimitMotor()->m_damping = 1.0; // default 1.0
		constraint->getTranslationalLimitMotor()->m_limitSoftness = 0.7; // default 0.7
		constraint->getTranslationalLimitMotor()->m_restitution = 0.5; // default 0.5
		for( i=3; i<6; i++ ){
			constraint->getRotationalLimitMotor( i )->m_damping = 1.0; // default 1.0
			constraint->getRotationalLimitMotor( i )->m_limitSoftness = 0.5; // default 0.5
			constraint->getRotationalLimitMotor( i )->m_bounce = 0.0; // default 0.0
		}
		
		//constraint->setLimit( 1, 1.0, 0.0 ); // linear y, this is the actual dof of interest
		
		constraint->setDbgDrawSize( 2.0 );
		
		data->jointFeedback = new btJointFeedback;
		constraint->setJointFeedback( data->jointFeedback );
		
		m_dynamicsWorld->addConstraint( constraint, true );
	}
	
	m_guiHelper->autogenerateGraphicsObjects( m_dynamicsWorld );
}

void Problem1::renderScene(){
	CommonRigidBodyBase::renderScene();
}

void Problem1::resetCamera(){
	float dist = 3;
	float pitch = -15;
	float yaw = 52;
	float targetPos[3] = { 0.0, 1.0, 0.0 };
	m_guiHelper->resetCamera(dist,yaw,pitch,targetPos[0],targetPos[1],targetPos[2]);
}

void Problem1::stepSimulation( float deltaTime ){
	m_dynamicsWorld->stepSimulation( deltaTime );
	
	btTransform transform;
	data->motionState->getWorldTransform( transform );
	
	printf( "appliedForce=%g linearVelocity=%g position=%g\n",
		data->jointFeedback->m_appliedForceBodyA.getY(),
		data->bodyPlatform->getLinearVelocity().getY(),
		transform.getOrigin().getY() );
	
	data->canvasTimeSeries->insertDataAtCurrentTime( data->jointFeedback->m_appliedForceBodyA.getY(), 0, true );
	data->canvasTimeSeries->insertDataAtCurrentTime( data->bodyPlatform->getLinearVelocity().getY(), 1, true );
	data->canvasTimeSeries->insertDataAtCurrentTime( transform.getOrigin().getY(), 2, true );
	data->canvasTimeSeries->nextTick();
}

CommonExampleInterface *Problem1CreateFunc( CommonExampleOptions &options ){
	return new Problem1( options.m_guiHelper );
}

B3_STANDALONE_EXAMPLE(Problem1CreateFunc)
